-----------------------------------------------------------------------------------------------
-- Client Lua Script for NPCBubbles
-- Written by Xan the Dragon // Gyro Egan
-- I don't care if you take any code from this and use it. Judging by the next line, NCSoft *technically* could but I don't think they will either.
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"

------------------------------
--CORE BACKEND FUNCTIONALITY--
------------------------------

local XNPCBubbles = {}
local BCHAT_CHANNEL = nil				-- Stores the channel instance used by NPCs.
local BCHAT_NAME = "Speech"				-- The display name of the channel. Must abide by channel name creation rules, and is thus strictly a-z, case insensitive only.
local BCHAT_ID = 255					-- The channel ID. This is the internal variable used by the game, and is responsible for why circles have /1, /2 - the number is the channel ID. This value can be changed to anything.
local MAX_DIST_DEFAULT = 36				-- The default value for the maximum distance that NPCs display their chats.
local MAX_DIST = MAX_DIST_DEFAULT		-- The current value for the maximum display distance, which is set to be the default value by default (legend has it that the very purpose of a 'default' is what we observe here!)
local MESSAGE_CACHE = {}				-- Stores messages recently said by any units.
local DISPLAY_SRC_CHANNEL = true		-- If this is true, the name of the NPC will have the source channel appended to it, if applicable. For instance, a message posted in NPC Say will show: [Speech] [NPC Say] UnitName: Message

function XNPCBubbles:new(extendObject)
    extendObject = extendObject or {}
    setmetatable(extendObject, self)
    self.__index = self
	return extendObject
end

---------------------
--UTILITY FUNCTIONS--
---------------------
--Attempts to locate the custom NPC speech channel if it hasn't been found already.
--If it finds the channel, it will set the global BCHAT_CHANNEL variable to store said channel.
--This function will not return any data. This is done because our interest is in setting the global variable BCHAT_CHANNEL to the channel.
--The global variable is used for the mod, of course, so it would be much easier to do what I am doing now than doing BCHAT_CHANNEL = FindAndSetChannel().
local function FindAndSetChannel()
	if BCHAT_CHANNEL ~= nil then
		return
	end
	for Index, Channel in pairs(ChatSystemLib.GetChannels()) do
		if Channel:GetName() == BCHAT_NAME then
			BCHAT_CHANNEL = Channel
			return --Why break when you can save 0.0000000000001 seconds by returning? Clock cycles are precious, you know!
			-- ...Does that even save time? probably not. asm would just say it exits the function either way. lol.
			-- I love it when I talk to myself like someone else is here, but I'm really just a lonely programmer in my basement.
		end
	end
end

--Function name is self explanatory but I should describe it anyway for good practice..
--Finds the distance from the input Unit to the current player. The Unit that's passed in should be the NPC.
local function DistanceToUnit(unit)
	local Me = GameLib.GetPlayerUnit()
	return (Vector3.New(Me:GetPosition()) - Vector3.New(unit:GetPosition())):Length()
	--You know, I've never seen anyone use this method before. I find that quite strange!
	--Most people write their own function that calculates it manually.
	--Hm. Food for thought.
end

--I don't like this method, but I suppose it's the only way.
local function IsCached(MessageText)
	for Index, Message in pairs(MESSAGE_CACHE) do
		if Message == MessageText then
			return true
		end
	end
	return false
end

local function PushToCache(Text)
	table.insert(MESSAGE_CACHE, 1, Text) --Push it to the start of the table.
	--10 should be more than enough cached messages. I am doing a cache array because I'm insecure about fast chats. /shrug
	if #MESSAGE_CACHE > 10 then
		table.remove(MESSAGE_CACHE, 11)
	end
end

--Temp, for debug. This is what helped me fix the issue on 243 with not realizing the messageInfo.arMessageSegments[1].strText code block, namely that random first index. [1].
function PrintTableContents(Table, Prefix)
	local Prefix = Prefix or "BASE->"
	for Index, Item in pairs(Table) do
		if type(Item) == "table" then
			PrintTableContents(Item, Prefix .. "[" .. tostring(Index) .. "|" .. tostring(Item) .. "]->")
		else
			Print(Prefix .. tostring(Index) .. "|" .. tostring(Item))
		end
	end
end 

--------------------------------
--[[INITIALIZATION FUNCTIONS]]--
--------------------------------
function XNPCBubbles:PreInit()
    Apollo.RegisterAddon(self, false, "", {})
end

function XNPCBubbles:OnLoad() 
	--Does this function have to be named OnLoad? I don't see where it's called within this lua file, so I suppose Apollo does that from its side.
	--Unfortunately this XML initialization line couldn't go. Originally it set some "self.xmlDoc" value but storing it would be useless since I never use UI.
	--This line has to stay since it is the only way to determine of the mod loaded + is ready to run. At least I think. There's no formal confirmation aside from the fact that it doesn't load when I remove it and replace it with simply calling the run function.
	
	--To do: Find any other ways of telling when the mod is ready, don't use the xml hook.
	self.xmlDoc = XmlDoc.CreateFromFile("NPCBubbles.xml")
	self.xmlDoc:RegisterCallback("Run", self)
end

--Data persistence saving
function XNPCBubbles:OnSave(DataLevel)
	if DataLevel == GameLib.CodeEnumAddonSaveLevel.Character then
		return {MAX_DIST, DISPLAY_SRC_CHANNEL}
	end
end

--Data persistence loading
function XNPCBubbles:OnRestore(DataLevel, LoadedData)
	if DataLevel == GameLib.CodeEnumAddonSaveLevel.Character then
		MAX_DIST = LoadedData[1] or MAX_DIST
		DISPLAY_SRC_CHANNEL = LoadedData[2]
		if DISPLAY_SRC_CHANNEL == nil then
			DISPLAY_SRC_CHANNEL = true --I can't do what I did for max distance because it will always pick the default (true) even if it's false, because that'd be (false or true) - the same outcome as (nil or true)
		end
	end
end

function XNPCBubbles:Run()
	if (self.xmlDoc == nil) or (not self.xmlDoc:IsLoaded()) then return end --uwot
	--Xan stop writing such horribly formatted conditions. What's wrong with you?
	
	--Covered above.
	FindAndSetChannel()
	local Found = BCHAT_CHANNEL ~= nil
	
	if not Found then
		--Channel wasn't located, so we'll have to create it, then let the player know it was added.
		--This will only call when it loads for the first time.
		
		ChatSystemLib.JoinChannel(BCHAT_NAME, "", BCHAT_ID) --Register the chat channel, let the player know.
		Print("[NPCBubbles] : The \"" .. BCHAT_NAME .. "\" channel has been registered. All NPC bubble chats will be sent here. Use {/bubbledist <number>} to set the maximum distance, or use {/bubbledist reset} to reset the value.")
		
		--Remember, the channel was NOT found. As a result, we need to call FindAndSet again, since that'll find (and set) the new channel that was just created.
		FindAndSetChannel()
		--Note that ^ is not guaranteed to set the value. In order to remedy this, I have it call this function if the channel is missing after an NPC chats.
		--...I think this is because the JoinChannel method is asynchronous to setting the actual channel list, so it's there but it just hasn't been added by the time we hit the right line.
	else
		--If it already exists just display the generic startup message.
		Print("[NPCBubbles] : NPCBubbles is active! Use {/bubbledist} to view commands for setting your preferences for this mod.")
	end
	--Display the warning. This should be done no matter what.
	Print("[NPCBubbles] : It is *strongly* recommended that you disable NPC Say in the list of channels you can see, since double posts may occur.")
	Apollo.RegisterEventHandler("UnitTextBubbleCreate", "OnChatBubbleCreated", self)
	Apollo.RegisterEventHandler("ChatMessage", "OnChatMessageAdded", self)
	Apollo.RegisterSlashCommand("bubbledist", "OnCommandRun", self)
	Apollo.RegisterSlashCommand("bubblesrc", "OnCommandRun", self)
end

-------------------
--[[EVENT HOOKS]]--
-------------------
function XNPCBubbles:OnCommandRun(cmd, argtext)
	if cmd and cmd:lower() == "bubbledist" then
		if argtext ~= nil and argtext ~= "" then
			local number = tonumber(argtext)
			if number ~= nil then
				MAX_DIST = math.min(256, math.max(number, 0))
				if number > 256 or number < 0 then
					Print("[WARN] The value " .. number .. " is out of bounds! It has been clamped to " .. MAX_DIST)
				else
					Print("The maximum bubble display distance has been set to " .. MAX_DIST)
				end
			else
				if argtext == "reset" then
					Print("The maximum bubble display distance has been reset to "..MAX_DIST_DEFAULT.." (Was " .. MAX_DIST .. ")")
					MAX_DIST = MAX_DIST_DEFAULT
				else
					Print("Invalid input. Expected a number or \"reset\".")
				end
			end
		else
			--No arguments supplied.
			--To do: NOT use a shitty command system. UI is annoying so I'll throw that into a SoonTM
			Print("NPCBubbles Help --")
			Print("/bubbledist <V> - Put any number where <V> is to set the maximum distance for bubbles. Any NPCs that use generic chatter (bubble chats) within this range will have their messages posted in the Speech channel")
			Print("/bubblesrc 1 - Use this command to enable displaying the source channel for any npc chats. For instance, if an NPC says something in NPC Say, the message in Speech will display: [Speech] [NPC Say] Name: Message")
			Print("/bubblesrc 0 - Use this command to disable the mode explained above.\n")
			Print("YOUR CURRENT MAXIMUM DISTANCE VALUE: " .. MAX_DIST .. ".")
			Print("You can reset it to the default value (" .. MAX_DIST_DEFAULT .. ") by saying /bubbledist reset")
		end
	elseif cmd and cmd:lower() == "bubblesrc" then
		--DISPLAY_SRC_CHANNEL
		if argtext ~= nil and argtext ~= "" then
			local number = tonumber(argtext)
			if number ~= nil then
				if number == 1 then
					Print("Message source display is now on.")
					DISPLAY_SRC_CHANNEL = true
				elseif number == 0 then
					Print("Message source display is now off.")
					DISPLAY_SRC_CHANNEL = false
				else
					Print("Invalid argument. Please use 0 or 1 (off or on, respectively)")
				end
			else
				if argtext == "reset" then
					Print("The maximum bubble display distance has been reset to "..MAX_DIST_DEFAULT.." (Was " .. MAX_DIST .. ")")
					MAX_DIST = MAX_DIST_DEFAULT
				else
					Print("Invalid input. Expected a number or \"reset\".")
				end
			end
		end
	end
end

function XNPCBubbles:PostToSpeech(Name, Text, ShowBubble, SourceChannel)
	if ShowBubble ~= true then
		ShowBubble = false --protect from nil stuff.
	end
	local Prefix = ""
	if SourceChannel then
		Prefix = "[" .. SourceChannel .. "] "
	end
	--The explanation for this block here is done in OnChatBubbleCreated
	if not BCHAT_CHANNEL then
		FindAndSetChannel()
		if not BCHAT_CHANNEL then
			return
		end
	end
	--Checking is done. Append the channel, if needed:
	if DISPLAY_SRC_CHANNEL then
		Name = Prefix .. Name --Fun fact: During development I had "Name .. Prefix". Then I called myself stupid for using prefix as a suffix. Look at me go!
		--Am I insane for talking to myself through comments?
	end
	
	--Checking complete. Final step:
	BCHAT_CHANNEL:Post(Text, Name, nil, ShowBubble)
end

function XNPCBubbles:OnChatMessageAdded(targetChannel, messageInfo)
	--ChatMessage event
	if targetChannel:GetName() == "NPC Say" or targetChannel:GetName() == "NPC Yell" or targetChannel:GetName() == "NPC Whisper" then
		--PrintTableContents(messageInfo)
		local name = messageInfo.unitSource:GetName()
		local text = messageInfo.arMessageSegments[1].strText --This took me like way too long. Docs said arMessageSegments was an array. NOT an array with another array as its first index. Carbine, you make me very very angry sometimes.
		if messageInfo.bShowChatBubble then
			PushToCache(text) --We do show the chat bubble by default. Cache this.
			--I was originally worried about a desync but the bubble created event has to fire after the chat event, as the creation of the bubble relies on it too. No desyncs!
			XNPCBubbles:PostToSpeech(name, text, false, targetChannel:GetName())
		else
			PushToCache(text) --We'll be reposting a chat with the bubble, so we cache it here to avoid it double posting.
			XNPCBubbles:PostToSpeech(name, text, true, targetChannel:GetName())
		end
	end
end

function XNPCBubbles:OnChatBubbleCreated(unitSpeaker, text, shouldBubbleSkipFade, bubbleMaxRange)
	--UnitTextBubbleCreate event.
	
	--New: Check for if this chat is cached to avoid reposting.
	--Reposting only lets stale memes live and be stolen >:(
	if IsCached(text) then
		return
	end
	
	--First check: Channel state. There's a chance that it won't load in time.
	if not BCHAT_CHANNEL then
		--Not registered, try to do so again.
		FindAndSetChannel()
		if not BCHAT_CHANNEL then
			--Originally I was going to display a warning here but it's useless since it'll load when it should, really. 
			--The only reason it wouldn't load would be if the player removed the channel themselves, in which case they want it to not display.
			--Even so, they should just disable the mod. ¯\_(•.•)_/¯
			return
		end
	end

	--Second check: Sent by an NPC?
	--EDIT: There's also a strange glitch causing the speaker's GetType method to be nil, and I mean the method itself, not the return value.
	--NOTE: Checking if the GetType index is nil or not doesn't work. Methods indexed with : cannot be treated normally (putting "if unitSpeaker:GetType ~= nil then" will throw a format error) 
	--The last option we have on our side is to do a protected call. Wish this wasn't the case but meh.
	--Edit: I was a big dumhed and forgot that lua allows for indexing methods with a dot.
	--....nvm the have it under a metatable.
	
	if not unitSpeaker then
		--Must've been a ghost talking.
		--Spooooooooooooky
		return
	end
	
	local IsNPC = false
	local SuccessfulCheck, ErrText = pcall(function ()
		if unitSpeaker:GetType() == "NonPlayer" then
			IsNPC = true
		end
	end)
	--if (not IsNPC) or (not SuccessfulCheck) then
	if not IsNPC then -- Do I keep this condition, ignoring IsNPC? EDIT: Yes because it'd still be false if SuccessfulCheck was false.
		return
	end
	
	--Third check: Distance.
	local Distance = DistanceToUnit(unitSpeaker)
	if Distance > MAX_DIST then
		return
	end
	
	--Checking complete. Final step:
	XNPCBubbles:PostToSpeech(unitSpeaker:GetName(), text, false)
end

local XNPCBubblesInst = XNPCBubbles:new()
XNPCBubblesInst:PreInit()
